mod codegen;
mod dynconst;
mod env;
mod intrinsics;
mod labeled_builder;
mod locs;
mod parser;
mod semant;
mod ssa;
mod types;

use std::fmt;
use std::path::Path;

use juno_scheduler::FunctionMap;

pub enum JunoVerify {
    None,
    JunoOpts,
    AllPasses,
}
impl JunoVerify {
    pub fn verify(&self) -> bool {
        match self {
            JunoVerify::None => false,
            _ => true,
        }
    }
    pub fn verify_all(&self) -> bool {
        match self {
            JunoVerify::AllPasses => true,
            _ => false,
        }
    }
}

pub enum JunoSchedule {
    None,
    DefaultSchedule,
    Schedule(String),
}

macro_rules! add_verified_pass {
    ($pm:ident, $verify:ident, $pass:ident) => {
        $pm.add_pass(hercules_opt::pass::Pass::$pass);
        if $verify.verify() || $verify.verify_all() {
            $pm.add_pass(hercules_opt::pass::Pass::Verify);
        }
    };
}
macro_rules! add_pass {
    ($pm:ident, $verify:ident, $pass:ident) => {
        $pm.add_pass(hercules_opt::pass::Pass::$pass);
        if $verify.verify_all() {
            $pm.add_pass(hercules_opt::pass::Pass::Verify);
        }
    };
}

pub enum ErrorMessage {
    SemanticError(semant::ErrorMessages),
    SchedulingError(String),
}

impl fmt::Display for ErrorMessage {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match self {
            ErrorMessage::SemanticError(errs) => {
                for err in errs {
                    write!(f, "{}\n", err)?;
                }
            }
            ErrorMessage::SchedulingError(msg) => {
                write!(f, "{}\n", msg)?;
            }
        }
        Ok(())
    }
}

pub fn compile(
    src_file: String,
    verify: JunoVerify,
    x_dot: bool,
    schedule: JunoSchedule,
    output_dir: String,
) -> Result<(), ErrorMessage> {
    let src_file_path = Path::new(&src_file);
    let module_name = String::from(src_file_path.file_stem().unwrap().to_str().unwrap());

    let prg = match semant::parse_and_analyze(src_file) {
        Ok(prg) => prg,
        Err(msg) => {
            return Err(ErrorMessage::SemanticError(msg));
        }
    };
    let (module, func_info) = codegen::codegen_program(prg);

    compile_ir(
        module,
        Some(func_info),
        verify,
        x_dot,
        schedule,
        output_dir,
        module_name,
    )
}

pub fn compile_ir(
    module: hercules_ir::ir::Module,
    func_info: Option<FunctionMap>,
    verify: JunoVerify,
    x_dot: bool,
    schedule: JunoSchedule,
    output_dir: String,
    module_name: String,
) -> Result<(), ErrorMessage> {
    let mut pm = match schedule {
        JunoSchedule::None => hercules_opt::pass::PassManager::new(module),
        _ => todo!(),
        /*
        JunoSchedule::DefaultSchedule => {
            let mut pm = hercules_opt::pass::PassManager::new(module);
            pm.make_plans();
            pm
        }
        JunoSchedule::Schedule(file) => {
            let Some(func_info) = func_info else {
                return Err(ErrorMessage::SchedulingError(
                    "Cannot schedule, no function information provided".to_string(),
                ));
            };

            match juno_scheduler::schedule(&module, func_info, file) {
                Ok(plans) => {
                    let mut pm = hercules_opt::pass::PassManager::new(module);
                    pm.set_plans(plans);
                    pm
                }
                Err(msg) => {
                    return Err(ErrorMessage::SchedulingError(msg));
                }
            }
        }
        */
    };
    if verify.verify() || verify.verify_all() {
        pm.add_pass(hercules_opt::pass::Pass::Verify);
    }
    add_verified_pass!(pm, verify, GVN);
    add_verified_pass!(pm, verify, PhiElim);
    add_pass!(pm, verify, DCE);
    if x_dot {
        pm.add_pass(hercules_opt::pass::Pass::Xdot(true));
    }
    add_pass!(pm, verify, Inline);
    if x_dot {
        pm.add_pass(hercules_opt::pass::Pass::Xdot(true));
    }
    // Inlining may make some functions uncalled, so run this pass.
    // In general, this should always be run after inlining.
    add_pass!(pm, verify, DeleteUncalled);
    if x_dot {
        pm.add_pass(hercules_opt::pass::Pass::Xdot(true));
    }
    // Run SROA pretty early (though after inlining which can make SROA more effective) so that
    // CCP, GVN, etc. can work on the result of SROA
    add_pass!(pm, verify, InterproceduralSROA);
    add_pass!(pm, verify, SROA);
    // We run phi-elim again because SROA can introduce new phis that might be able to be
    // simplified
    add_verified_pass!(pm, verify, PhiElim);
    add_pass!(pm, verify, DCE);
    if x_dot {
        pm.add_pass(hercules_opt::pass::Pass::Xdot(true));
    }
    add_pass!(pm, verify, CCP);
    add_pass!(pm, verify, DCE);
    add_pass!(pm, verify, GVN);
    add_pass!(pm, verify, DCE);
    add_pass!(pm, verify, WritePredication);
    add_pass!(pm, verify, PhiElim);
    add_pass!(pm, verify, DCE);
    add_pass!(pm, verify, Predication);
    add_pass!(pm, verify, DCE);
    add_pass!(pm, verify, CCP);
    add_pass!(pm, verify, DCE);
    add_pass!(pm, verify, GVN);
    add_pass!(pm, verify, DCE);
    if x_dot {
        pm.add_pass(hercules_opt::pass::Pass::Xdot(true));
    }
    //add_pass!(pm, verify, Forkify);
    //add_pass!(pm, verify, ForkGuardElim);
    add_verified_pass!(pm, verify, DCE);
    add_pass!(pm, verify, ForkSplit);
    add_pass!(pm, verify, Unforkify);
    add_pass!(pm, verify, GVN);
    add_verified_pass!(pm, verify, DCE);
    add_pass!(pm, verify, DCE);
    add_pass!(pm, verify, Outline);
    add_pass!(pm, verify, InterproceduralSROA);
    add_pass!(pm, verify, SROA);
    add_pass!(pm, verify, InferSchedules);
    add_verified_pass!(pm, verify, DCE);
    if x_dot {
        pm.add_pass(hercules_opt::pass::Pass::Xdot(true));
    }

    add_pass!(pm, verify, GCM);
    add_verified_pass!(pm, verify, DCE);
    add_pass!(pm, verify, FloatCollections);
    add_pass!(pm, verify, GCM);
    pm.add_pass(hercules_opt::pass::Pass::Codegen(output_dir, module_name));
    pm.run_passes();

    Ok(())
}
